package info.markhillman.SOMColorSorter;

import java.awt.*;

/**
 * Class: Som
 * Description: This class will use a Self-Organising Map
 * in order to sort a set of data into identifiable clusters.
 * The Som contains an array of SOMNodes which will be used to
 * produce the detection grid
 * Created by Mark on 08/01/2016.
 */
public class Som {

    //Array of SOMNodes
    //private int neighbourDistance;
    private SomNode[][] nodes;
    private float neighbourRadius;
    private float learningRate;
    private int inputSize;

    public Som(int nodeNum, int inputSize) {

        //Create the nodes according the parameters passed in
        this.inputSize = inputSize;
        nodes = new SomNode[nodeNum][nodeNum];
        for (int i = 0; i < nodes.length; i++) {
            for (int j = 0; j < nodes.length; j++) {
                nodes[i][j] = new SomNode(inputSize);
            }
        }

        //Set the neighbourDistance as the size of the grid
        //neighbourDistance = nodeNum;
        neighbourRadius = nodeNum / 2;

        //Set the learning rate
        learningRate = 0.1f;
    }

    // Reset the SOMNodes weights
    public void reset() {
        for (int i = 0; i < nodes.length; i++) {
            for (int j = 0; j < nodes.length; j++) {
                nodes[i][j].randomiseWeights();
            }
        }
        neighbourRadius = nodes.length / 2;
        learningRate = 0.1f;
    }

    //This will train the Som for 1 epoch
    //This function takes the array of all the training data
    public void train(float[][] inputDataArray) {

        //Get the random index
        int index = (int)Math.floor(Math.random() * inputDataArray.length);

        //Get the data
        float[] inputData = inputDataArray[index];

        //Get the best node for the data
        Point best = getBestNode(inputData);

        //Update the Som based upon the best node
        updateNeighbourhood(inputData, best.x, best.y);
    }

    //This will find the best SomNode and return it
    private Point getBestNode(float[] inputData) {

        //Run through the list of nodes and check which on has the
        //smallest distance value

        //Get the first value
        float minDistance = nodes[0][0].calculateDistance(inputData);
        Point best = new Point(0, 0);

        //Iterate to find best node
        for (int i = 1; i < nodes.length; i++) {
            for (int j = 0; j < nodes[i].length; j++) {
                float distance = nodes[i][j].calculateDistance(inputData);
                if (distance < minDistance) {
                    minDistance = distance;
                    best = new Point(i, j);
                }
            }
        }

        return best;
    }

    //This will alter a SomNode and its neighbours
    private void updateNeighbourhood(float[] inputData, int bestX, int bestY) {

        /*
        //Update the best node
        nodes[bestX][bestY].update(inputData, learningRate);
        int i = bestX, j = bestY;
        int x, y;

        //Iterate through the neighbours
        for (int k = 0; k < neighbourDistance; k++) {
            for(x = Math.max(0, i - k); x <= Math.min(i + k, nodes.length - 1); x++) {
                for(y = Math.max(0, j - k); y <= Math.min(j + k, nodes[0].length - 1); y++) {
                    if(x != i || y != j) {

                        //Update node x, y
                        nodes[x][y].update(inputData, learningRate);
                    }
                }
            }
        }
        neighbourDistance--;
        neighbourDistance = (neighbourDistance < 5) ? 5 : neighbourDistance;
        learningRate = learningRate - (learningRate / 1000);
        */

        //Loop through the relevant SOMNodes
        float radiusSqaured = neighbourRadius * neighbourRadius;
        for (int i = Math.max(0, bestX - (int)neighbourRadius); i < Math.min(nodes.length, bestX + neighbourRadius); i++) {
            for (int j = Math.max(0, bestY - (int)neighbourRadius); j < Math.min(nodes.length, bestY + neighbourRadius); j++) {

                //Calculate the distance from the best
                int xDist = bestX - i;
                int yDist = bestY - j;
                int distanceSquared = (xDist * xDist) + (yDist * yDist);

                //Check the distance is within the radius
                if (distanceSquared < radiusSqaured) {

                    //Calculate the dampening
                    float dampening = (float)Math.exp(-(distanceSquared / radiusSqaured * 3));

                    //Update the nodes
                    nodes[i][j].update(inputData, learningRate * dampening);
                }
            }
        }

        //Reduce the neighbour radius and learning rate
        //Arbitrary reductions
        learningRate *= 0.9999f;
        neighbourRadius *= 0.9995f;
        //System.out.println("Neighbour Radius: " + neighbourRadius + ", Learning rate: " + learningRate);
    }

    //Get the nodes array
    public SomNode[][] getNodes() {
        return nodes;
    }
}
