package info.markhillman.SOMColorSorter;

import javafx.application.Platform;
import javafx.concurrent.Service;
import javafx.concurrent.Task;
import javafx.concurrent.Worker;
import javafx.fxml.FXML;
import javafx.fxml.Initializable;
import javafx.scene.canvas.Canvas;
import javafx.scene.canvas.GraphicsContext;
import javafx.scene.control.Button;
import javafx.scene.control.ColorPicker;
import javafx.scene.control.TextField;
import javafx.scene.image.ImageView;
import javafx.scene.image.PixelWriter;
import javafx.scene.image.WritableImage;
import javafx.scene.layout.AnchorPane;
import javafx.scene.paint.Color;

import java.awt.*;
import java.net.URL;
import java.util.ResourceBundle;

/**
 * Class:
 * Description:
 * Created by Mark on 19/03/2016.
 */
public class SomDisplayControllerFx implements Initializable {

    private float[][] inputDataArray;
    private int somSize;
    private Som som;
    private SomNode[][] nodes;

    @FXML private AnchorPane imagePane;
    @FXML private Button resetButton;
    @FXML private Button trainButton;
    @FXML private Button stopButton;
    @FXML private Button findColorButton;
    @FXML private Button randomColorButton;
    @FXML private TextField epochField;
    @FXML private Canvas imageCanvas;
    @FXML private ColorPicker colorPicker;
    private GraphicsContext context;
    private WritableImage image;
    private PixelWriter pixels;
    private int radius = 15;
    private Task<Void> trainer;

    @Override
    public void initialize(URL location, ResourceBundle resources) {

        // Create the normal data
        // Randomly make 20 different colors
        inputDataArray = new float[40][3];
        for (int i = 0; i < inputDataArray.length; i++) {
            for (int j = 0; j < inputDataArray[i].length; j++) {
                inputDataArray[i][j] = (float)Math.random();
            }
        }

        // Create the Som with somSize*somSize array and 3 values (RGB)
        somSize = 400;
        som = new Som(somSize, inputDataArray[0].length);
        nodes = som.getNodes();

        // Setup image canvas
        imageCanvas.setHeight(somSize + (radius * 2));
        imageCanvas.setWidth(somSize + (radius * 2));
        context = imageCanvas.getGraphicsContext2D();
        image = new WritableImage(somSize, somSize);
        pixels = image.getPixelWriter();

        // Add button listeners
        resetButton.setOnAction(event -> reset());
        trainButton.setOnAction(event -> train());
        stopButton.setOnAction(event -> stopTraining());
        findColorButton.setOnAction(event -> findColor());
        randomColorButton.setOnAction(event -> randomColor());

        reset();
    }

    // This will start a training thread and update the image
    public void train() {
        String epochString = new String(epochField.getCharacters().toString());
        try {

            //Cancel old trainers
            stopTraining();

            //Create new trainer
            int epochs = Integer.parseInt(epochString);
            trainer = new Task<Void>() {
                @Override public Void call() {
                    for (int i = 0; i < epochs; i++) {
                        if (isCancelled())
                            return null;
                        trainSom(1);
                    }
                    return null;
                }
            };
            new Thread(trainer).start();

        } catch (NumberFormatException e) {
            System.out.println(e);
        }
    }

    // Reset the Som data to randomly initialized data
    public void reset() {
        stopTraining();
        som.reset();
        updateImage();
    }

    // This will stop current training if the som has a training thread running
    public void stopTraining() {
        if (trainer != null && trainer.isRunning()) {
            trainer.cancel();
        }
    }

    // Find the closet weight to the selected color
    public void findColor(){

        stopTraining();

        float distance = 3;
        Point closest = new Point();
        Color color = colorPicker.getValue();
        for (int i = 0; i < nodes.length; i++) {
            for (int j = 0; j < nodes.length; j++) {
                float currentDistance = 0;
                float[] rgbArray = nodes[i][j].getWeights();
                currentDistance += Math.abs(color.getRed() - rgbArray[0]);
                currentDistance += Math.abs(color.getGreen() - rgbArray[1]);
                currentDistance += Math.abs(color.getBlue() - rgbArray[2]);

                if (currentDistance < distance) {
                    closest = new Point(i, j);
                    distance = currentDistance;
                }
            }
        }

        // Draw circle
        context.setFill(color);
        context.setStroke(Color.BLACK);
        context.setLineWidth(3);
        context.fillOval(closest.getX(), closest.getY(), radius * 2, radius * 2);
        context.strokeOval(closest.getX(), closest.getY(), radius * 2, radius * 2);
    }

    public void randomColor() {
        Color color = new Color(Math.random(), Math.random(), Math.random(), 1);
        colorPicker.setValue(color);
    }

    // This will train the Som for the specified number of epochs
    private synchronized void trainSom(int epochs) {
        for (int i = 0; i < epochs; i++) {
            som.train(inputDataArray);
            updateImage();
        }
    }

    // This will redraw the image and add it to the window
    private synchronized void updateImage() {
        for (int i = 0; i < somSize; i++) {
            for (int j = 0; j < somSize; j++) {
                float[] rgbArray = nodes[i][j].getWeights();
                pixels.setColor(i, j, new Color(rgbArray[0], rgbArray[1], rgbArray[2], 1));
            }
        }
        context.clearRect(0, 0, somSize + (radius * 2), somSize + (radius * 2));
        context.drawImage(image, radius, radius);
    }
}
